//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "ScriptExec.h"
#include "Log.h"
#include "LuaState.h"
#include "lua.hpp"
#include <Clipbrd.hpp>
#include <string>
#include <assert.h>
#include <time.h>
#include <map>


namespace {
	std::map<lua_State*, ScriptExec*> contexts;

	ScriptExec* GetContext(lua_State* L)
	{
		std::map<lua_State*, ScriptExec*>::iterator it;
		it = contexts.find(L);
		if (it == contexts.end())
		{
			assert(!"Unregistered lua_State!");
			return NULL;
		}
		return it->second;
	}

	std::map<AnsiString, AnsiString> variables;

	long timediff(clock_t t1, clock_t t2) {
		long elapsed;
		elapsed = static_cast<long>(((double)t2 - t1) / CLOCKS_PER_SEC * 1000);
		return elapsed;
	}
}

int ScriptExec::LuaPrint(lua_State *L)
{
	int nArgs = lua_gettop(L);
	int i;
	lua_getglobal(L, "tostring");
	std::string ret;//this is where we will dump the output
	//make sure you start at 1 *NOT* 0
	for(i = 1; i <= nArgs; i++)
	{
		const char *s;
		lua_pushvalue(L, -1);
		lua_pushvalue(L, i);
		lua_call(L, 1, 1);
		s = lua_tostring(L, -1);
		if(s == NULL)
			return luaL_error(L, LUA_QL("tostring") " must return a string to ", LUA_QL("print"));

		if(i > 1) ret.append("\t");

		ret.append(s);
		lua_pop(L, 1);
	};
	//ret.append("\n");
	GetContext(L)->onAddOutputText(ret.c_str());
	return 0;
};

int ScriptExec::LuaError( lua_State *L )
{
	const char* str = lua_tostring( L, -1 );
	lua_pop(L, 1);
	//send string to console
	GetContext(L)->onAddOutputText(str);
	return 0;
}

int ScriptExec::l_ShowMessage( lua_State* L )
{
	AnsiString msg = lua_tostring(L, -1);
	MessageBox(NULL, msg.c_str(), "Lua message", MB_ICONINFORMATION);
	lua_pushnumber( L, 0 );
	return 1;
}

int ScriptExec::l_Sleep(lua_State* L)
{
	int ret = 0;
	int time = lua_tounsigned(L, -1);
	clock_t t1 = clock();
	long elapsed = 0;
	ScriptExec* context = GetContext(L);
	if (context->breakReq)
	{
		ret = -1;
	}
	else
	{
		while (elapsed < time)
		{
			if (context->breakReq)
			{
				ret = -1;
				break;
			}
			Application->ProcessMessages();
			Sleep(1);
			elapsed = timediff(t1, clock());
			if (context->breakReq)
			{
				ret = -1;
				break;
			}
		}
	}
	lua_pushnumber(L, ret);
	return 1;					// number of return values
}

/** Beep(frequency, time_ms)
	\note blocking and not interruptable
*/
int ScriptExec::l_Beep(lua_State *L)
{
	/*
	To refer to elements in the stack, the API uses indices. The first element in the stack (that is, the element that was pushed first) has index 1, the next one has index 2, and so on. We can also access elements using the top of the stack as our reference, using negative indices. In that case, -1 refers to the element at the top (that is, the last element pushed), -2 to the previous element, and so on. For instance, the call lua_tostring(L, -1) returns the value at the top of the stack as a string.
	*/
	int freq = lua_tounsigned(L, 1);
	int time = lua_tounsigned(L, 2);
	Beep(freq, time);
	return 0;
}

int ScriptExec::l_GetClipboardText( lua_State* L )
{
	AnsiString text = Clipboard()->AsText;
	lua_pushstring( L, text.c_str() );
	return 1;
}

int ScriptExec::l_SetClipboardText( lua_State* L )
{
	const char* str = lua_tostring( L, -1 );
	if (str == NULL)
	{
		LOG("Lua error: str == NULL\n");
		return 0;
	}
	Clipboard()->AsText = str;
	return 0;
}

int ScriptExec::l_Call(lua_State* L)
{
	const char* str = lua_tostring( L, -1 );
	if (str == NULL)
	{
		LOG("Lua error: str == NULL\n");
		return 0;
	}
	GetContext(L)->onCall(str);
	return 0;
}

int ScriptExec::l_Hangup(lua_State* L)
{
	GetContext(L)->onHangup();
	return 0;
}

int ScriptExec::l_SetDial(lua_State* L)
{
	const char* str = lua_tostring( L, -1 );
	if (str == NULL)
	{
		LOG("Lua error: str == NULL\n");
		return 0;
	}
	GetContext(L)->onSetDial(str);
	return 0;
}

int ScriptExec::l_GetDial(lua_State* L)
{
	std::string num = GetContext(L)->onGetDial();
	lua_pushstring( L, num.c_str() );
	return 1;
}

int ScriptExec::l_SetInitialCallTarget(lua_State* L)
{
	const char* str = lua_tostring( L, -1 );
	if (str == NULL)
	{
		LOG("Lua error: str == NULL\n");
		return 0;
	}
	GetContext(L)->onSetInitialCallTarget(str);
	return 0;
}

int ScriptExec::l_GetInitialCallTarget(lua_State* L)
{
	std::string num = GetContext(L)->onGetInitialCallTarget();
	lua_pushstring( L, num.c_str() );
	return 1;
}

int ScriptExec::l_SwitchAudioSource(lua_State* L)
{
	//  The first element in the stack (that is, the element that was pushed first) has index 1, the next one has index 2, and so on.
	const char* mod = lua_tostring( L, 1 );
	if (mod == NULL)
	{
		LOG("Lua error: mod == NULL\n");
		return 0;
	}
	const char* dev = lua_tostring( L, 2 );
	if (dev == NULL)
	{
		LOG("Lua error: dev == NULL\n");
		return 0;
	}
	GetContext(L)->onSwitchAudioSource(mod, dev);
	return 0;
}

int ScriptExec::l_SendDtmf(lua_State* L)
{
	const char* str = lua_tostring( L, -1 );
	if (str == NULL)
	{
		LOG("Lua error: str == NULL\n");
		return 0;
	}
	GetContext(L)->onSendDtmf(str);
	return 0;
}

int ScriptExec::l_BlindTransfer(lua_State* L)
{
	const char* str = lua_tostring( L, -1 );
	if (str == NULL)
	{
        LOG("Lua BlindTransfer error: str == NULL\n");
		return 0;
	}
	GetContext(L)->onBlindTransfer(str);
	return 0;
}

int ScriptExec::l_GetCallState(lua_State* L)
{
	int state = GetContext(L)->onGetCallState();
	lua_pushinteger( L, state );
	return 1;
}

int ScriptExec::l_IsCallIncoming(lua_State* L)
{
	int incoming = GetContext(L)->onIsCallIncoming();
	lua_pushinteger( L, incoming );
	return 1;
}

int ScriptExec::l_GetCallPeer(lua_State* L)
{
	std::string num = GetContext(L)->onGetCallPeer();
	lua_pushstring( L, num.c_str() );
	return 1;
}

int ScriptExec::l_GetStreamingState(lua_State* L)
{
	int state = GetContext(L)->onGetStreamingState();
	lua_pushinteger( L, state );
	return 1;
}

int ScriptExec::l_SetVariable(lua_State* L)
{
	const char* name = lua_tostring( L, 1 );
	if (name == NULL)
	{
		LOG("Lua error: name == NULL\n");
		return 0;
	}
	const char* value = lua_tostring( L, 2 );
	if (value == NULL)
	{
		LOG("Lua error: value == NULL\n");
		return 0;
	}
	variables[name] = value;
	return 0;
}

int ScriptExec::l_GetVariable(lua_State* L)
{
	const char* name = lua_tostring( L, 1 );
	if (name == NULL)
	{
		LOG("Lua error: name == NULL\n");
		return 0;
	}
	std::map<AnsiString, AnsiString>::iterator it;
	it = variables.find(name);
	if (it != variables.end())
	{
		// returning two variables, second one informs if variable is found
		lua_pushstring( L, it->second.c_str() );
		lua_pushinteger( L, 1 );
	}
	else
	{
		lua_pushstring( L, "" );
		lua_pushinteger( L, 0 );
	}
	return 2;
}

int ScriptExec::l_ClearVariable(lua_State* L)
{
	const char* name = lua_tostring( L, 1 );
	if (name == NULL)
	{
		LOG("Lua error: name == NULL\n");
		return 0;
	}	
	std::map<AnsiString, AnsiString>::iterator it;
	it = variables.find(name);
	if (it != variables.end())
	{
        variables.erase(it);
	}
	return 0;
}

int ScriptExec::l_ClearAllVariables(lua_State* L)
{
	variables.clear();
	return 0;
}

int ScriptExec::l_ShellExecute(lua_State* L)
{
	const char* verb = lua_tostring( L, 1 );
	const char* file = lua_tostring( L, 2 );
	const char* parameters = lua_tostring( L, 3 );
	const char* directory = lua_tostring( L, 4 );
	int showCmd = lua_tointeger( L, 5 );
	int ret = reinterpret_cast<int>(ShellExecute(NULL, verb, file, parameters, directory, showCmd));
    lua_pushinteger(L, ret);
	return 1;
}


ScriptExec::ScriptExec(
	CallbackAddOutputText onAddOutputText,
	CallbackCall onCall,
	CallbackHangup onHangup,
	CallbackGetDial onGetDial,
	CallbackSetDial onSetDial,
	CallbackSwitchAudioSource onSwitchAudioSource,
	CallbackSendDtmf onSendDtmf,
	CallbackBlindTransfer onBlindTransfer,
	CallbackGetCallState onGetCallState,
	CallbackIsCallIncoming onIsCallIncoming,
	CallbackGetCallPeer onGetCallPeer,
	CallbackGetStreamingState onGetStreamingState,
	CallbackGetInitialCallTarget onGetInitialCallTarget,
	CallbackSetInitialCallTarget onSetInitialCallTarget
	):
	onAddOutputText(onAddOutputText),
	onCall(onCall),
	onHangup(onHangup),
	onGetDial(onGetDial),
	onSetDial(onSetDial),
	onSwitchAudioSource(onSwitchAudioSource),
	onSendDtmf(onSendDtmf),
	onBlindTransfer(onBlindTransfer),
	onGetCallState(onGetCallState),
	onIsCallIncoming(onIsCallIncoming),
	onGetCallPeer(onGetCallPeer),
	onGetStreamingState(onGetStreamingState),
	onGetInitialCallTarget(onGetInitialCallTarget),
	onSetInitialCallTarget(onSetInitialCallTarget),

	breakReq(false),
	running(false)
{
	assert(onAddOutputText && onCall && onHangup && onGetDial && onSetDial &&
		onSwitchAudioSource && onSendDtmf && onBlindTransfer && onGetCallState &&
		onIsCallIncoming && onGetCallPeer && onGetStreamingState &&
		onGetInitialCallTarget && onSetInitialCallTarget);
}

ScriptExec::~ScriptExec()
{
}

void ScriptExec::Run(const char* script)
{
	breakReq = false;
	running = true;
	Lua_State L;
	luaL_openlibs(L);
	contexts[L] = this;

	lua_register(L, "_ALERT", LuaError );
	lua_register(L, "print", LuaPrint );
	lua_register(L, "ShowMessage", l_ShowMessage);
	lua_register(L, "Sleep", l_Sleep);
	lua_register(L, "Beep", l_Beep);
	lua_register(L, "GetClipboardText", l_GetClipboardText);
	lua_register(L, "SetClipboardText", l_SetClipboardText);
	lua_register(L, "Call", l_Call);
	lua_register(L, "Hangup", l_Hangup);
	lua_register(L, "GetDial", l_GetDial);
	lua_register(L, "SetDial", l_SetDial);
	lua_register(L, "SwitchAudioSource", l_SwitchAudioSource);
	lua_register(L, "SendDtmf", l_SendDtmf);
	lua_register(L, "BlindTransfer", l_BlindTransfer);
	lua_register(L, "GetCallState", l_GetCallState);
	lua_register(L, "IsCallIncoming", l_IsCallIncoming);
	lua_register(L, "GetCallPeer", l_GetCallPeer);
	lua_register(L, "GetStreamingState", l_GetStreamingState);
	lua_register(L, "SetVariable", l_SetVariable);
	lua_register(L, "GetVariable", l_GetVariable);
	lua_register(L, "ClearVariable", l_ClearVariable);
	lua_register(L, "ClearAllVariables", l_ClearAllVariables);
	lua_register(L, "GetInitialCallTarget", l_GetInitialCallTarget);
	lua_register(L, "SetInitialCallTarget", l_SetInitialCallTarget);
    lua_register(L, "ShellExecute", l_ShellExecute);


	int res = luaL_dostring(L, script);
	if(res != 0)
	{
		AnsiString txt;
		txt.sprintf("Execution error:\n%s", lua_tostring(L, -1));
		MessageBox(NULL, txt.c_str(), "Lua", MB_ICONINFORMATION);
	}
	running = false;

	std::map<lua_State*, ScriptExec*>::iterator it;
	it = contexts.find(L);
	assert(it != contexts.end());
	contexts.erase(it);
}

void ScriptExec::Break(void)
{
	breakReq = true;
}

